#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <libintl.h> 
#include <newt.h>
#include <popt.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <locale.h> 

#if defined(__sparc__)
#include <sys/ioctl.h>
#include <sys/sysmacros.h>
#include <fcntl.h>

#define	KEYMAPDIRSUN	"/lib/kbd/keymaps/sun"
#define	KEYMAPDIRPC	"/lib/kbd/keymaps/i386"
#define KEYMAPDIR	keymapdir
unsigned char *keymapdir;
#else
#define	KEYMAPDIR	"/lib/kbd/keymaps/i386"
#endif
unsigned int keyboardtype = -1;
#define KBDTYPE_SUN	0
#define KBDTYPE_PC	1
#define KBDTYPE_SERIAL	2

#define _(String) gettext((String)) 

static char * progName;

static int getKeyboardConfig(char ** table, int useBack) {
    FILE * f;
    char buf[255];
    char start[255];
    char * end;
    int num = 0;
    int rc;
    int i, j;
    int item;
    char **maps = NULL;
    DIR *mapdir;
    struct dirent *ent;
    char * fullDirStack[50];
    char ** dirStack = fullDirStack;
    char * fullName;
    char * currDir;
    struct stat sb;

    *dirStack++ = NULL;
    *dirStack = strdup(KEYMAPDIR);
    
    while (*dirStack) {
	currDir = *dirStack--;

	mapdir = opendir(currDir);
	if (!mapdir) {
	    *table = _("cannot scan /lib/kbd/keymaps for keyboard maps");
	    return -1;
	}

	errno = 0;
	ent = readdir(mapdir);
	if (errno) {
	    *table = _("cannot read first /lib/kbd/keymaps keyboard map");
	    return -1;
	}

	while (ent) {
	    char *s, *t;

	    strcpy(start, ent->d_name);
	    if ((t=strstr(start, ".kmap")) || (t=strstr(start, ".map"))) {
		*t = '\0';
		maps = (char **) realloc(maps, (num+1)*sizeof(char *));
		maps[num] = malloc(strlen(start)+5);
		strcpy(maps[num], start);
		num++;
	    } else if (ent->d_name[0] != '.') {
		fullName = malloc(strlen(ent->d_name) + strlen(currDir) + 2);
		sprintf(fullName, "%s/%s", currDir, ent->d_name);
		stat(fullName, &sb);
		if (S_ISDIR(sb.st_mode)) {
		    *(++dirStack) = fullName;
		} else {
		    free(fullName);
		}
	    }

	    ent = readdir(mapdir);
	    if (errno) {
		*table = _("cannot read next /lib/kbd/keymaps keyboard map");
		free(maps);
		return -1;
	    }
	}

	maps = (char **) realloc(maps, (num+1)*sizeof(char *));
	maps[num] = NULL;

	free(currDir);
    }

    closedir(mapdir);

    for (i = 0; i < num - 1; i++) {
	for (j = 0; j < num - 1; j++) {
	    int result;
	    char *tmp;

	    result = strcmp(maps[j], maps[j+1]);
            if (result > 0) {
                tmp = maps[j];
                maps[j] = maps[j+1];
                maps[j+1] = tmp;
            }
        }
    }

    for (i = 0; i < num - 1; i++) {
	if (!strcmp(maps[i], *table)) {
	    item = i;
	    break;
	}
    }

    rc = newtWinMenu(_("Configure Keyboard"), 
		     _("What type of keyboard do you have?"), 40, 0, 10,
		     8, maps, &item, _("Ok"), 
		     useBack ? _("Back") : _("Cancel"), NULL);

    if (rc != 2) {
	*table = maps[item];
	rc = 0;
    } else {
	*table = NULL;
	rc = 1;
    }
 
    free(maps);

    return rc;
}


/* reads /etc/sysconfig/kbd, which just has KEYTABLE=<path> and optionally KEYBOARDTYPE=<sun|pc> */
/* double quotes are stripped, #'s denote comments */
static int readConfigFile(char ** table) {
    FILE *f;
    char *s, *se;
    char buf[250];
    int line = 0;

    f = fopen("/etc/sysconfig/keyboard", "r");
    if (!f) {
	if (errno == ENOENT) {
	    return 0;
	}

	fprintf(stderr, _("%s: cannot open /etc/sysconfig/keyboard: %s\n"),
		progName, strerror(errno));
	return 1;
    }

    while ((s = fgets(buf, sizeof(buf) - 1, f)) != NULL) {
	line++;

	/* skipping leading spaces */
	while (*s && isspace(*s)) s++;
	if (!*s) continue;

	/* skip comments */
	if (*s == '#') continue;

	/* cut off trailing spaces and \n */
	se = s + strlen(s) - 2;
	while (isspace(*se)) se--;
	se++;
	*se = '\0';	
	se--;
	
	if (!strncmp("KEYTABLE=", s, 9)) {
	    s += 9;
	    if (*s == '"' && *se != '"') {
		fprintf(stderr, _("%s: mismatched quotes on line %d in "
			"/etc/sysconfig/keyboard\n"), progName, line);
		fclose(f);
		return 1;
	    } else if (*s == '"') {
		s++;
		*se = '\0';
		se--;
	    }

	    s = basename(s);
	    if (!strncmp(se-3, ".map", 4)) {
		se -= 3;
		*se = '\0';
	    }

	    *table = strdup(s);
	} else if (!strncmp("KEYBOARDTYPE=", s, 13)) {
	    s += 13;
	    if (*s == '"' && *se != '"') {
		fprintf(stderr, _("%s: mismatched quotes on line %d in "
			"/etc/sysconfig/keyboard\n"), progName, line);
		fclose(f);
		return 1;
	    } else if (*s == '"') {
		s++;
		*se = '\0';
		se--;
	    }
#ifdef __sparc__	    
	    if (!strcasecmp (s, "sun")) {
	    	if (keyboardtype == -1) keyboardtype = KBDTYPE_SUN;
	    } else
#endif
	    if (!strcasecmp (s, "pc")) {
	    	if (keyboardtype == -1) keyboardtype = KBDTYPE_PC;
	    } else {
#ifdef __sparc__
	    	fprintf(stderr, _("%s: KEYBOARDTYPE has to be either \"sun\" or \"pc\" on line %d in /etc/sysconfig/keyboard\n"), progName, line);
#else
	    	fprintf(stderr, _("%s: KEYBOARDTYPE has to be \"pc\" on line %d in /etc/sysconfig/keyboard\n"), progName, line);
#endif
	    	fclose(f);
	    	return 1;
	    }
	} else {
	    fprintf(stderr, _("%s: line %d unexpected in "
		    "/etc/sysconfig/keyboard\n"), progName, line);
	    fclose(f);
	    return 1;
	}
    }

    return 0;
}

#ifdef __sparc__
int checkSunKeyboard() {
	int fd, fdstd = 0;
	char twelve = 12;
	struct stat s;
	char buf[20];

	/* Test for serial console */
	for (fd = 0; fd <= 2; fd++) {
	    sprintf (buf, "/proc/self/fd/%d", fd);
	    if (readlink (buf, buf, 4096) == 12 &&
		!strncmp (buf, "/dev/console", 12)) {
		fdstd = 1;
		break;
	    }
	}
	if (!fdstd) {
	    fd = open("/dev/console", O_RDONLY);
	    if (fd < 0) return -1;
	}
	if (ioctl (fd, TIOCLINUX, &twelve) < 0) {
		if (!fdstd) close(fd);
		return KBDTYPE_SERIAL;
	}
	if (!fdstd) close(fd);
	fd = open("/dev/kbd", O_RDONLY);
	if (fd < 0)
		return KBDTYPE_PC;
    	close(fd);
    	return KBDTYPE_SUN;
}
#endif

int main(int argc, char ** argv) {
    char buf[1024];
    int i, rc;
    FILE * f;
    char * table = NULL;
    char * origTable;
    char * roottext[128];
    newtComponent form;
    int back = 0;
    int test = 0;
    poptContext optCon;
    struct poptOption options[] = {
            { "back", '\0', 0, &back, 0 },
            { "test", '\0', 0, &test, 0 },
            { 0, 0, 0, 0, 0 }
        };

    progName = argv[0];

    setlocale(LC_ALL, ""); 
    bindtextdomain("kbdconfig","/usr/share/locale"); 
    textdomain("kbdconfig"); 

    optCon = poptGetContext("kbdconfig", argc, (const char **)argv, options,0);
    poptReadDefaultConfig(optCon, 1);

    if ((rc = poptGetNextOpt(optCon)) < -1) {
	fprintf(stderr, _("%s: bad argument %s: %s\n"), progName,
		poptBadOption(optCon, POPT_BADOPTION_NOALIAS), 
		poptStrerror(rc));
	return 2;
    }

    if (poptGetArg(optCon)) {
	fprintf(stderr, _("%s: unexpected argument\n"), progName);
	return 2;
    }

    poptFreeContext(optCon);

    if (!test && getuid()) {
	fprintf(stderr, _("%s: can only be run as root\n"), progName);
	exit(2);
    }

#ifdef __sparc__
    keyboardtype = checkSunKeyboard();
#else
    keyboardtype = KBDTYPE_PC;
#endif 

    if (readConfigFile(&table)) {
	fprintf(stderr, _("%s: critical error reading /etc/sysconfig/keyboard"),
		progName);
	exit(2);
    }

#ifdef __sparc__
    if (keyboardtype == -1)
    	keyboardtype = KBDTYPE_SUN;
    if (keyboardtype == KBDTYPE_PC)
    	keymapdir = KEYMAPDIRPC;
    else
    	keymapdir = KEYMAPDIRSUN;
    if (keyboardtype == KBDTYPE_SUN && !table)
        table = "sunkeymap";
    else
#endif  

    if (!table)
	table = "us";

    origTable = strdup(table);

    newtInit();
    newtCls();
	
    newtPushHelpLine(_(" <Tab>/<Alt-Tab> between elements   |  <Space> selects   |  <F12> next screen"));
    snprintf((char *)roottext,128, _("kbdconfig %s - (C) 1999-2001 Red Hat, Inc."), VERSION);
	
    newtDrawRootText(0, 0, (const char *)roottext);

    if (keyboardtype == KBDTYPE_SERIAL) {
	newtWinMessage(_("Serial console"), _("Ok"), _("No keymap will be loaded on serial console"));
	newtFinished();
	if (test)
	    exit (0);
	if (unlink ("/etc/sysconfig/keyboard") < 0 && errno != ENOENT) {
	    fprintf(stderr, _("%s: could not remove /etc/sysconfig/keyboard: %s"), 
		    progName, strerror(errno));
	    exit (1);
	}
	if (!access("/etc/sysconfig/console/",W_OK)) {
	    if (unlink ("/etc/sysconfig/console/default.kmap") < 0 && errno != ENOENT) {
		fprintf(stderr, _("%s: could not remove /etc/sysconfig/console/default.kmap: %s"), 
			progName, strerror(errno));
		exit (1);
	    }
	}
	exit (0);
    }
    
    if (getKeyboardConfig(&table, back)) {
	/* cancelled */
	newtFinished();

	if (table) {
	    fprintf(stderr, "%s: %s\n", progName, table);
	    exit(2);
	}

	exit(1);
    }

    if (test) {
	    newtFinished();
	    return 0;
    }

    f = fopen("/etc/sysconfig/keyboard", "w");
    if (!f) {
	newtFinished();
	fprintf(stderr, _("%s: failed to make /etc/sysconfig/keyboard: %s"), 
		progName, strerror(errno));
	exit(2);
    }

    if (fprintf(f, "KEYTABLE=\"%s\"\n", table) < 0) {
	newtFinished();
	fprintf(stderr, 
		_("%s: failed to write to /etc/sysconfig/keyboard: %s"), 
		progName, strerror(errno));
	unlink("/etc/sysconfig/keyboard");
	exit(2);
    }

#ifdef __sparc__
    if (fprintf(f, "KEYBOARDTYPE=\"%s\"\n", keyboardtype == KBDTYPE_PC ? "pc" : "sun") < 0) {
	newtFinished();
	fprintf(stderr, 
		_("%s: failed to write to /etc/sysconfig/keyboard: %s"), 
		progName, strerror(errno));
	unlink("/etc/sysconfig/keyboard");
	exit(2);
    }
#endif 
    fclose(f);
    if (strcmp(origTable, table)) {
	sprintf(buf, "loadkeys \"%s\" 2> /dev/null", table);
	system(buf);
	newtWinMessage(_("Keymap Loaded"), _("Ok"), 
		       _("The new keymap has been loaded."));
    }

    newtFinished();
    /* copy default keyboard map */
    if (!access("/etc/sysconfig/console/",W_OK)) {
	sprintf(buf, "dumpkeys >/etc/sysconfig/console/default.kmap 2>/dev/null");
	system(buf);
    }
    return 0;
}
